#####################################################################################################################################################
#---------------------------------------------------------------------------------------------------------------------------------------------------#
#	- This file is a part of the VocalTractLab Python module PyVTL, see https://github.com/paul-krug/VocalTractLab
#---------------------------------------------------------------------------------------------------------------------------------------------------#
#
#	- Copyright (C) 2021, Paul Konstantin Krug, Dresden, Germany
#	- https://github.com/paul-krug/VocalTractLab
#	- Author: Paul Konstantin Krug, TU Dresden
#
#	- License info:
#
#		This program is free software: you can redistribute it and/or modify
#		it under the terms of the GNU General Public License as published by
#		the Free Software Foundation, either version 3 of the License, or
#		(at your option) any later version.
#		
#		This program is distributed in the hope that it will be useful,
#		but WITHOUT ANY WARRANTY; without even the implied warranty of
#		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#		GNU General Public License for more details.
#		
#		You should have received a copy of the GNU General Public License
#		along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#---------------------------------------------------------------------------------------------------------------------------------------------------#
#####################################################################################################################################################
#---------------------------------------------------------------------------------------------------------------------------------------------------#

#---------------------------------------------------------------------------------------------------------------------------------------------------#
#####################################################################################################################################################
#---------------------------------------------------------------------------------------------------------------------------------------------------#
# Load essential packages:
import VocalTractLab.VocalTractLabApi as vtl
import VocalTractLab.plotting_tools as PT
from VocalTractLab.plotting_tools import finalize_plot
from VocalTractLab.plotting_tools import get_plot
from VocalTractLab.plotting_tools import get_plot_limits
import librosa
import numpy as np
import matplotlib.pyplot as plt
from scipy.signal import find_peaks
#---------------------------------------------------------------------------------------------------------------------------------------------------#
#####################################################################################################################################################




#---------------------------------------------------------------------------------------------------------------------------------------------------#
def multiple_formatter( denominator=2, number=np.pi, latex='\\pi' ):
	def gcd(a, b):
		while b:
			a, b = b, a%b
		return a
	def _multiple_formatter(x, pos):
		den = denominator
		num = np.int(np.rint(den*x/number))
		com = gcd(num,den)
		(num,den) = (int(num/com),int(den/com))
		if den==1:
			if num==0:
				return r'$0$'
			if num==1:
				 return r'$%s$'%latex
			elif num==-1:
				return r'$-%s$'%latex
			else:
				return r'$%s%s$'%(num,latex)
		else:
			if num==1:
				return r'$\frac{%s}{%s}$'%(latex,den)
			elif num==-1:
				return r'$\frac{-%s}{%s}$'%(latex,den)
			else:
				return r'$\frac{%s%s}{%s}$'%(num,latex,den)
	return _multiple_formatter
#---------------------------------------------------------------------------------------------------------------------------------------------------#
class Multiple:
	def __init__(self, denominator=2, number=np.pi, latex='\\pi'):
		self.denominator = denominator
		self.number = number
		self.latex = latex
	def locator(self):
		return plt.MultipleLocator(self.number / self.denominator)
	def formatter(self):
		return plt.FuncFormatter(multiple_formatter(self.denominator, self.number, self.latex))
#---------------------------------------------------------------------------------------------------------------------------------------------------#



#####################################################################################################################################################
class Transfer_Function():
	def __init__( self, 
		          magnitude_spectrum: np.ndarray,
		          phase_spectrum: np.ndarray,
		          n_spectrum_samples: int,
		          name: str = 'transfer_function'
		          ):
		if not isinstance( n_spectrum_samples, int ):
			raise ValueError( 'n_spectrum_samples must be an integer and should be a power of 2! Passed type is: {}'.format( type( n_spectrum_samples ) ) )
		self.constants = vtl.get_constants()
		self.delta_frequency = self.constants[ 'samplerate_audio' ] / n_spectrum_samples
		max_bin = round( n_spectrum_samples / self.delta_frequency )
		self.n_spectrum_samples = n_spectrum_samples
		if isinstance( magnitude_spectrum, np.ndarray ):
			self.magnitude_spectrum = magnitude_spectrum[ : max_bin ]
		else:
			self.magnitude_spectrum = None
		if isinstance( phase_spectrum, np.ndarray ):
			self.phase_spectrum = phase_spectrum[ : max_bin ]
		else:
			self.phase_spectrum = None
		self.data = dict( frequency = self.magnitude_spectrum, phase = self.phase_spectrum )
		self.formants = self.get_formants()
		self.f1, self.f2, self.f3, self.f4 = self.formants
		return
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	def get_formants( self, peak_distance = 1, sr = 44100 ):
		sr = self.constants[ 'samplerate_audio' ]
		peaks, _ = find_peaks( self.magnitude_spectrum, distance = peak_distance )
		peaks = [ peak * sr/self.n_spectrum_samples for peak in peaks ]
		while peaks[ 0 ] < 100:
			del peaks[ 0 ]
		if len( peaks ) < 4:
			peaks.extend( [ None for _ in range( 0, 4 - len( peaks ) ) ] )
		elif len( peaks ) > 4:
			peaks = peaks[ : 4 ]
		return peaks
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	def plot( self,
		      parameters = [ 'frequency', 'phase' ],
		      plot_formants = True,
		      axs: list = None,
		      plot_kwargs: list = [ dict( color = 'navy' ), dict( color = 'darkorange' ) ],
		      **kwargs,
		      ): #, scale = 'dB' ):
		figure, axs = get_plot( n_rows = len( parameters ), axs = axs )
		for index, parameter in enumerate( parameters ):
			if parameter == 'frequency':
				y = librosa.amplitude_to_db( self.data[ parameter ] )
				continuities = [ slice( 0, len(y) ) ]
				y_title = 'Intensity [dB]'
				#_min = np.min( y )
				#_max = np.max( y )
				#axs[ index ].set( ylim = [ _min - 0.1 * np.abs( _max - _min ), _max + 0.1 * np.abs( _max - _min ) ] )
				axs[ index ].set( ylim = get_plot_limits( y ) )
				axs[ index ].locator_params( axis = 'y', nbins = 4 )
			elif parameter == 'phase':
				continuities = []
				y = self.data[ parameter ]
				tmp_idx = 0
				for idx in range( 0, len(y) - 1 ):
					if np.abs( y[idx] - y[idx+1] ) > 1.5:
						continuities.append( slice( tmp_idx, idx+1 ) )
						tmp_idx = idx + 1
				if tmp_idx != len( y ):
					continuities.append( slice( tmp_idx, len( y ) ) )

				y = self.data[ parameter ]
				y_title = 'Phase'
				axs[ index ].yaxis.set_major_locator(plt.MultipleLocator(np.pi / 2))
				#axs[ index ].yaxis.set_minor_locator(plt.MultipleLocator(np.pi / 12))
				axs[ index ].yaxis.set_major_formatter(plt.FuncFormatter(multiple_formatter()))
				axs[ index ].set( ylim = [ -3.76, 3.76 ] )
			else:
				raise ValueError( 'parameters must be frequency and/or phase! Passed values are: {}'.format( parameters ) )
			x = np.arange( 0, self.n_spectrum_samples, self.delta_frequency )
			for _slice in continuities:
				axs[ index ].plot( x[ _slice ], y[ _slice ], **plot_kwargs[ index ] )
			axs[ index ].set( ylabel = y_title )
		plt.xlabel( 'Frequency [Hz]' )
		if plot_formants:
			for formant in self.formants:
				for ax in axs:
					ax.axvline( formant, color = 'gray', ls = '--' )
		for ax in axs:
			ax.label_outer()
		finalize_plot( figure, axs, **kwargs )
		return axs
#---------------------------------------------------------------------------------------------------------------------------------------------------#


#####################################################################################################################################################
#---------------------------------------------------------------------------------------------------------------------------------------------------#
#	- This file is a part of the VocalTractLab Python module PyVTL, see https://github.com/paul-krug/VocalTractLab
#---------------------------------------------------------------------------------------------------------------------------------------------------#
#
#	- Copyright (C) 2021, Paul Konstantin Krug, Dresden, Germany
#	- https://github.com/paul-krug/VocalTractLab
#	- Author: Paul Konstantin Krug, TU Dresden
#
#	- License info:
#
#		This program is free software: you can redistribute it and/or modify
#		it under the terms of the GNU General Public License as published by
#		the Free Software Foundation, either version 3 of the License, or
#		(at your option) any later version.
#		
#		This program is distributed in the hope that it will be useful,
#		but WITHOUT ANY WARRANTY; without even the implied warranty of
#		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#		GNU General Public License for more details.
#		
#		You should have received a copy of the GNU General Public License
#		along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#---------------------------------------------------------------------------------------------------------------------------------------------------#
#####################################################################################################################################################
#---------------------------------------------------------------------------------------------------------------------------------------------------#

#---------------------------------------------------------------------------------------------------------------------------------------------------#
#####################################################################################################################################################
#---------------------------------------------------------------------------------------------------------------------------------------------------#
# Load essential packages:
import warnings
import numpy as np
import matplotlib.pyplot as plt
from VocalTractLab.function_tools import check_if_list_is_valid
from VocalTractLab.function_tools import make_output_path
from VocalTractLab.function_tools import is_iterable


hist_kwargs = dict( bins = 50, histtype = 'step', density = False )

state_plot_kwargs = dict( HX = dict( color = '#02010e' ),
                          HY = dict( color = '#02010e' ),
                          JX = dict( color = 'navy' ),
                          JA = dict( color = 'navy' ),
                          LP = dict( color = 'rebeccapurple' ),
                          LD = dict( color = 'rebeccapurple' ),
                          VS = dict( color = 'darkmagenta' ),
                          VO = dict( color = 'darkmagenta' ),
                          TCX = dict(  color = 'darkorange' ),
                          TCY = dict(  color = 'darkorange' ),
                          TTX = dict(  color = 'darkorange' ),
                          TTY = dict(  color = 'darkorange' ),
                          TBX = dict(  color = 'darkorange' ),
                          TBY = dict(  color = 'darkorange' ),
                          TRX = dict(  color = 'darkorange' ),
                          TRY = dict(  color = 'darkorange' ),
                          TS1 = dict(  color = 'darkorange' ),
                          TS2 = dict(  color = 'darkorange' ),
                          TS3 = dict(  color = 'darkorange' ),
                          F0 = dict( color = 'teal' ),
                          PR = dict( color = 'teal' ),
                          XB = dict( color = 'turquoise' ),
                          XT = dict( color = 'turquoise' ),
                          CA = dict( color = 'turquoise' ),
                          PL = dict( color = 'turquoise' ),
                          RA = dict( color = 'turquoise' ),
                          DP = dict( color = 'turquoise' ),
                          PS = dict( color = 'turquoise' ),
                          FL = dict( color = 'turquoise' ),
                          AS = dict( color = 'turquoise' ),
                          )

state_hist_kwargs = { key: { **val, **hist_kwargs } for key, val in state_plot_kwargs.items() } # In Python 3.9+: key: orig | extra

segment_plot_kwargs = dict( boundaries = dict( color = 'black' ),
                            phonemes = dict( color = 'black', ha = 'center', va = 'center' ),
                            )


def get_plot( n_rows = 1,
	          n_columns = 1,
	          axs = None, 
	          sharex = True,
	          gridspec_kw = {'hspace': 0},
	          **kwargs,
	          ):
	#if n_columns == 1:
	#	gridspec_kw = gridspec_kw = {'hspace': 0}
	#else:
	#	gridspec_kw = {}
	if axs == None:
		figure, axs = plt.subplots( n_rows, n_columns, figsize = (8, 4/3 * n_rows ), sharex = sharex, gridspec_kw = gridspec_kw )
	else:
		figure = None
	try:
		if len( axs ) != ( n_rows * n_columns ) and n_columns == 1:
			raise ValueError( 'Length of passed matplotlib.pyplot.axes ({}) list does not equal the number of parameters to plot ({}).'.format( len( axs ), n_rows * n_columns ) )
	except Exception:
		axs = [ axs ]
	return [ figure, np.array( axs ) ]


def get_plot_limits( data, offset = 0.1 ):
	_min = np.min( data )
	_max = np.max( data )
	if _min == _max:
		return [ _min - 0.75, _max + 0.75 ]
	else:
		return [ _min - offset * np.abs( _max - _min ), _max + offset * np.abs( _max - _min ) ]

def get_valid_tiers( parameters, container, types = str ):
	#container = check_if_list_is_valid( container, str )
	if parameters == None:
		valid_parameters = container
	else:
		parameters = check_if_list_is_valid( parameters, types )
		valid_parameters = []
		for parameter in parameters:
			if parameter in set( container ):
				valid_parameters.append( parameter )
			else:
				warnings.warn( 'The specified parameter: {} does not exist in the sequence, plotting skipped.'.format( parameter ) )
	return valid_parameters

def finalize_plot( figure = None,
	               axs = None,
	               hide_labels = True,
	               out_file_path = None,
	               overwrite = False,
	               show = True,
	               ):
	if hide_labels:
		for ax in axs.flatten():
			ax.label_outer()
	if ( out_file_path != None ) or show:
		if ( figure != None ) and is_iterable( axs ):
			figure.align_ylabels( axs[:] )
		plt.tight_layout()
		if out_file_path != None:
			out_file_path = make_output_path( None, out_file_path, overwrite )
			plt.savefig( out_file_path )
		if show:
			plt.show()
	return


#####################################################################################################################################################
#---------------------------------------------------------------------------------------------------------------------------------------------------#
#	- This file is a part of the VocalTractLab Python module PyVTL, see https://github.com/paul-krug/VocalTractLab
#---------------------------------------------------------------------------------------------------------------------------------------------------#
#
#	- Copyright (C) 2021, Paul Konstantin Krug, Dresden, Germany
#	- https://github.com/paul-krug/VocalTractLab
#	- Author: Paul Konstantin Krug, TU Dresden
#
#	- License info:
#
#		This program is free software: you can redistribute it and/or modify
#		it under the terms of the GNU General Public License as published by
#		the Free Software Foundation, either version 3 of the License, or
#		(at your option) any later version.
#		
#		This program is distributed in the hope that it will be useful,
#		but WITHOUT ANY WARRANTY; without even the implied warranty of
#		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#		GNU General Public License for more details.
#		
#		You should have received a copy of the GNU General Public License
#		along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#---------------------------------------------------------------------------------------------------------------------------------------------------#
#####################################################################################################################################################
#---------------------------------------------------------------------------------------------------------------------------------------------------#

#---------------------------------------------------------------------------------------------------------------------------------------------------#
#####################################################################################################################################################
#---------------------------------------------------------------------------------------------------------------------------------------------------#
# Load essential packages:
from VocalTractLab.g2p import text_to_phonemes
from VocalTractLab.function_tools import check_if_list_is_valid
from VocalTractLab.segment_sequence import Segment_Sequence
from VocalTractLab.VocalTractLabApi import gestural_score_to_audio
from VocalTractLab.VocalTractLabApi import segment_sequence_to_gestural_score



def text_to_speech( sentence_list, language = 'en', **kwargs ):
	sentence_list = check_if_list_is_valid( sentence_list, str )
	phonemes_list = text_to_phonemes( sentence_list, phoneme_style = 'vtl_sampa', language = language )
	segment_sequences = get_phone_durations( phonemes_list, method = 'naive', language = language )
	for index, seg in enumerate( segment_sequences ):
		seg.to_seg_file( 'tts_tmp_{}.seg'.format( index ) )
	segment_sequence_to_gestural_score( [ 'tts_tmp_{}.seg'.format( index ) for index, _ in enumerate( segment_sequences ) ] )
	audio_data = gestural_score_to_audio( [ 'tts_tmp_{}.ges'.format( index ) for index, _ in enumerate( segment_sequences ) ],
		                                  save_file = False,
		                                  normalize_audio = -1,
		                                  sr = 16000,
		                                  return_data = True,
		                                  )
	return audio_data
def tts( text, language = 'de', **kwargs ):
	return text_to_speech( text, language = language, **kwargs )

def get_phone_durations( phonemes_list, method = 'naive', language = 'en' ):
	segment_sequences = []
	for phonemes in phonemes_list:
		phonemes_flat = []
		for _list in phonemes:
			phonemes_flat += _list
		segment_sequence = Segment_Sequence( phonemes_flat, [ 0.2 for _ in phonemes_flat ] )
		segment_sequences.append( segment_sequence )
	return segment_sequences



#####################################################################################################################################################
#---------------------------------------------------------------------------------------------------------------------------------------------------#
#	- This file is a part of the VocalTractLab Python module PyVTL, see https://github.com/paul-krug/VocalTractLab
#---------------------------------------------------------------------------------------------------------------------------------------------------#
#
#	- Copyright (C) 2021, Paul Konstantin Krug, Dresden, Germany
#	- https://github.com/paul-krug/VocalTractLab
#	- Author: Paul Konstantin Krug, TU Dresden
#
#	- License info:
#
#		This program is free software: you can redistribute it and/or modify
#		it under the terms of the GNU General Public License as published by
#		the Free Software Foundation, either version 3 of the License, or
#		(at your option) any later version.
#		
#		This program is distributed in the hope that it will be useful,
#		but WITHOUT ANY WARRANTY; without even the implied warranty of
#		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#		GNU General Public License for more details.
#		
#		You should have received a copy of the GNU General Public License
#		along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#---------------------------------------------------------------------------------------------------------------------------------------------------#
#####################################################################################################################################################
#---------------------------------------------------------------------------------------------------------------------------------------------------#

#---------------------------------------------------------------------------------------------------------------------------------------------------#
#####################################################################################################################################################
#---------------------------------------------------------------------------------------------------------------------------------------------------#
# Load essential packages:
#cimport cVocalTractLabApi

import numpy as np
cimport numpy as np
import pandas as pd
#import ctypes
import os
import warnings
import time

from libcpp cimport bool
from libcpp.vector cimport vector  

import logging

logging.basicConfig()
log = logging.getLogger(__name__)
log.setLevel(logging.WARNING)

#import atexit

from cpython.pycapsule cimport *
#from dataclasses import dataclass
import utils_method as um
#from libcpp.vector cimport vector  
#from cpython cimport array
#from libc.stdlib cimport malloc, free
import matplotlib.pyplot as plt
from itertools import cycle

import VocalTractLab.targets as tg
from VocalTractLab.plotting_tools import get_plot_limits
#from VocalTractLab.targets import Target_Sequence









#---------------------------------------------------------------------------------------------------------------------------------------------------#
#####################################################################################################################################################
#
#	C++ API functions:
#
#####################################################################################################################################################
#---------------------------------------------------------------------------------------------------------------------------------------------------#
cdef extern from "Data.h":
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	cdef struct Sample:
		double time, value
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	cdef struct PitchTarget:
		double slope, offset, tau, duration
#---------------------------------------------------------------------------------------------------------------------------------------------------#
cdef extern from "TargetOptimizerApi.h":
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	cdef struct FitData:
		vector[ PitchTarget ] res_targets,
		vector[ double ] res_boundaries,
		vector[ Sample ] res_trajectory,
		vector[ double ] res_ftmp,
		double res_fmin,
		double res_rmse,
		double res_corr,
		double res_time,
		double res_onset,
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	FitData estimate_targets(	
							#double *arr_times,
							#double *arr_values,
							#double *arr_boundaries,
							vector[double] input_times,
							vector[double] input_values,
							vector[double] boundaries,
							#int init_bounds,

							double weight_slope,
							double weight_offset,
							double weight_tau,
							double weight_lambda,

							double delta_slope,
							double delta_offset,
							double delta_tau,
							double delta_boundary,

							double mean_slope,
							#double mean_offset,
							double mean_tau,

							int max_iterations,
							int max_cost_evaluations,
							double rho_end,

							bool use_early_stopping,
							double epsilon,
							int patience,
						);
#---------------------------------------------------------------------------------------------------------------------------------------------------#
#####################################################################################################################################################






#---------------------------------------------------------------------------------------------------------------------------------------------------#
class Time_Signal():
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	def __init__( self, input_times, input_values ):
		self.times = input_times
		self.values = input_values
		return
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	@classmethod
	def from_pitch_tier_file( cls, pitch_tier_file_path ):
		df_pitch_tier = pd.read_csv( pitch_tier_file_path , sep = '\t', skiprows= 3 , header = None )
		df_pitch_tier.columns = [ 'time', 'value' ]
		#print( df_pitch_tier )
		return cls( df_pitch_tier[ 'time' ].to_numpy(), df_pitch_tier[ 'value' ].to_numpy() )
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	@classmethod
	def from_csv( cls, csv_file_path, **kwargs ):
		df = pd.read_csv( csv_file_path , **kwargs )
		#df.columns = [ 'time', 'value' ]
		return cls( df[ 'time' ], df[ 'value' ] )
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	def get_samples( self, ):
		cdef vector[ Sample ] samples
		cdef Sample sample
		for time, value in zip( self.times, self.values ):
			sample.time = time
			sample.value = value
			samples.push_back( sample )
		return samples
#---------------------------------------------------------------------------------------------------------------------------------------------------#



#---------------------------------------------------------------------------------------------------------------------------------------------------#
class Fit_Result():
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	@um.autoargs()
	def __init__(
		self,
		in_times,
		in_values,
		in_boundaries,
		par_init_bounds,
		par_weight_slope,
		par_weight_offset,
		par_weight_tau,
		par_weight_lambda,
		par_delta_slope,
		par_delta_offset,
		par_delta_tau,
		par_delta_boundary,
		par_mean_slope,
		#par_mean_offset: float
		par_mean_tau,
		par_max_iterations,
		par_max_cost_evaluations,
		par_rho_end,
		par_use_early_stopping,
		par_epsilon,
		par_patience,
		out_targets,
		out_boundaries,
		out_trajectory,
		out_ftmp,
		out_fmin,
		out_rmse,
		out_corr,
		out_time,
		):
		pass
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	def plot( self, ):
		min_in_value = np.min( self.in_values )
		max_in_value = np.max( self.in_values ) 
		plt.scatter( self.in_times, self.in_values, c = 'darkorange' )
		plt.plot( self.out_trajectory[ :, 0 ], self.out_trajectory[ :, 1 ], color = 'navy' )
		for in_boundary, out_boundary in zip( self.in_boundaries, self.out_boundaries):
			plt.axvline( in_boundary, ymin = 0.95, color = 'black' )
			plt.axvline( out_boundary, ymax = 0.95, color = 'lightgray', ls = '--' )
		plt.ylim( [ min_in_value - 0.2 * np.abs( max_in_value - min_in_value ), max_in_value + 0.2 * np.abs( max_in_value - min_in_value ) ] )
		plt.show()
		return
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	def to_csv( self, csv_file_path ):
		return
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	def to_gestural_score( self, ges_file_path, ges_type = 'f0-gestures', unit = 'st' ): #Attention: neutral is always 0
		ges_file = open( 'ges_file_path', 'w' )
		ges_file.write( '<gestural_score>\n' )
		ges_file.write( '\t<gesture_sequence type="{}" unit="{}">\n'.format( ges_type, unit) )
		for target in self.out_targets:
			ges_file.write( '\t\t<gesture value="{}" slope="{}" duration_s="{}" time_constant_s="{}" neutral="0" />\n'.format(
				target.offset, target.slope, target.duration, target.tau ) )
		ges_file.write( '\t</gesture_sequence>\n' )
		ges_file.write( '</gestural_score>\n' )
		return
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	def to_pitch_tier( self, pitch_tier_file_path ):
		pitch_tier_file = open( 'pitch_tier_file_path', 'w' )
		pitch_tier_file.write( 'ooTextFile\n' )
		pitch_tier_file.write( 'PitchTier\n' )
		pitch_tier_file.write( '{} {} {}\n'.format( self.file_onset, self.file_offset, self.out_trajectory.shape[0] ) )
		for sample in self.out_trajectory:
			pitch_tier_file.write( '{}\t{}\n'.format( sample[0], sample[1] ) )
		return
#---------------------------------------------------------------------------------------------------------------------------------------------------#


#---------------------------------------------------------------------------------------------------------------------------------------------------#
class Fit_Window():
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	def __init__( self, times, values, boundaries, delta_boundary, queue_position, window_length = 3, hop_length = 1, **kwargs ):
		self.times = []
		self.values = []
		self.boundaries = boundaries[ queue_position : queue_position + window_length + 1 ] # Todo Update: hop_length * queue_pos....
		for time, value in zip( times, values ):
			if time >= self.boundaries[ 0 ] and time <= self.boundaries[ -1 ]:
				self.times.append( time )
				self.values.append( value )
		self.delta_boundary = delta_boundary
		self.queue_position = queue_position
		self.target_positions = range( queue_position + hop_length, queue_position + hop_length + window_length )
		self.window_length = window_length
		self.hop_lenth = hop_length
		self.fit_result = None
		self.kwargs = kwargs
		return
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	def fit( self, **kwargs ):
		self.fit_result = fit(
			times = self.times,
			values = self.values,
			boundaries = self.boundaries,
			delta_boundary = self.delta_boundary,
			**self.kwargs
			)
		return
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	def plot():
		return
#---------------------------------------------------------------------------------------------------------------------------------------------------#

#---------------------------------------------------------------------------------------------------------------------------------------------------#


#---------------------------------------------------------------------------------------------------------------------------------------------------#
class Sequential_Fit():
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	def __init__( self, windows, times, values ):
		self.windows = windows
		self.times = times
		self.values = values
		for window in self.windows:
			window.fit()
		return
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	#def get_optimized_boundaries( windows, window_length, hop_length ):
	#	boundary_sequence = []
	#	n_targets = windows[-1].target_positions[-1] + 1
	#	target_index = cycle( [ x for x in range(0, window_length) ] )
	#	for index in range( 0, hop_length +1 ):
	#		boundary_sequence.append( windows[0].fit_result.out_targets[ target_index[ index ] ].onset_time )
	#	for index in range( hop_length + 1, n_targets - 1 ):
	#		tmp_targets = []
	#		for window in windows:
	#			if ( index in window.target_positions ) and ( index != window.target_positions[0] ):
	#				tmp_targets.append( window.fit_result.out_targets[ target_index[ index ] ] )
	#		onset_time_mean = np.mean( [ x.onset_time for x in tmp_targets ] )
	#		boundary_sequence.append( onset_time_mean )
	#	for index in range( n_targets - 1, n_targets ):
	#		boundary_sequence.append( windows[0].fit_result.out_targets[ target_index[ index ] ].onset_time )
	#	boundary_sequence.append( windows[0].fit_result.out_targets[ target_index[ index ] ].offset_time )
	#	return boundary_sequence
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	#def get_optimized_targets( windows, hop_length ):
	#	n_windows = len( windows )
	#	target_list = []
	#	for window in windows:
	#		if window.queue_position == 0:
	#			target_list.extend( window.fit_result.out_targets[ : 1 + hop_length ] )
	#		elif window.queue_position < n_windows - 1:
	#			target_list.append( window.fit_result.out_targets[ hop_length ] )
	#		elif window.queue_position == n_windows - 1:
	#			target_list.extend( window.fit_result.out_targets[ 1 : ] )
	#	return target_list
#---------------------------------------------------------------------------------------------------------------------------------------------------#
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	def plot( self, ):
		figure, axs = plt.subplots( len( self.windows ), figsize = (12, 4/3 *len( self.windows ) ), sharex = True, gridspec_kw = {'hspace': 0} )
		low, high = get_plot_limits( self.values, 0.2 )
		for index, window in enumerate( self.windows ):
			for boundary in window.boundaries:
				for ax in axs:
					ax.axvline( boundary, color = 'gray', ls = '--', zorder = 2 )
				axs[ window.queue_position ].axvline( boundary, color = 'black', zorder = 3 )
			axs[ window.queue_position ].axvspan( 0, window.boundaries[0], alpha = 0.5, color = 'lightgray' )
			axs[ window.queue_position ].axvspan( window.boundaries[ -1 ], self.windows[ -1 ].boundaries[ -1 ], alpha = 0.5, color = 'lightgray' )
			axs[ window.queue_position ].scatter( self.times, self.values, color = 'gray', s=10, zorder = 2 )
			axs[ window.queue_position ].scatter( window.fit_result.in_times, window.fit_result.in_values, color = 'black', s=20, zorder = 3 )
			axs[ window.queue_position ].set_ylim( low, high )
			axs[ window.queue_position ].set_ylabel( '{}'.format( index + 1 ), color = (0,0,0,0) )
			axs[ window.queue_position ].set_yticks([])
			axs[ window.queue_position ].set_yticklabels([])
			if window.queue_position == 0:
				axs[ window.queue_position ].axvspan( window.boundaries[0], window.boundaries[-1], alpha = 0.5, color = 'lightgreen' )
			else:
				axs[ window.queue_position ].axvspan( window.boundaries[0], window.boundaries[1], alpha = 0.5, color = 'red' )
				axs[ window.queue_position ].scatter(
					window.boundaries[0] + 0.5 * (window.boundaries[1] - window.boundaries[0]),
					low + 0.5 * (high-low), 
					s=600, c='red', marker='x', zorder= 10 )
				axs[ window.queue_position ].axvspan( window.boundaries[1], window.boundaries[-1], alpha = 0.5, color = 'lightgreen' )



		plt.xlabel( 'Time [s]' )
		figure.text( 0.02, 0.6, 'Contour [arb]', va='center', rotation='vertical' )
		#figure.supylabel('Contour [arb]')
		for ax in axs:
			ax.label_outer()
		figure.align_ylabels( axs[:] )
		plt.tight_layout()
		plt.savefig( 'sequential_fit_principle.pdf' )
		plt.show()
		return
#---------------------------------------------------------------------------------------------------------------------------------------------------#

#---------------------------------------------------------------------------------------------------------------------------------------------------#




#---------------------------------------------------------------------------------------------------------------------------------------------------#
#####################################################################################################################################################
#
#	User functions:
#
#####################################################################################################################################################
#---------------------------------------------------------------------------------------------------------------------------------------------------#

def convert_hz_to_st( values ):
	return 12 * ( np.log( values ) / np.log( 2 ) )


#---------------------------------------------------------------------------------------------------------------------------------------------------#
# 		Single core functions
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def fit(	
			times,
			values,
			boundaries = [],
			init_bounds = 4,
			weight_slope = 1.0, 
			weight_offset = 0.5,
			weight_tau = 0.1, 
			weight_lambda = 0.01,
			delta_slope = 0.5,
			delta_offset = 1.0,
			delta_tau = 5,
			delta_boundary = 100.0,
			mean_slope = 0.0,
			#mean_offset = None,
			mean_tau = 20.0,
			max_iterations = 20,
			max_cost_evaluations = 1e6,
			rho_end = 1e-6,
			use_early_stopping = False,
			epsilon = 0.01,
			patience = None,
			):
	if patience == None:
		patience = -1
	if boundaries == [] and init_bounds != 0:
		step =  ( times[ -1 ] - times[ 0 ] ) / ( init_bounds - 1 )
		boundaries = np.array( [ times[ 0 ] + x * step for x in range( 0 , init_bounds ) ] )
		#print( boundaries )
		#print( len(boundaries ) )
	#if boundaries[ 0 ] > times[ 0 ]:
	#	boundaries[ 0 ] = times[ 0 ] - 0.01
	#if boundaries[ -1 ] < times[ -1 ]:
	#	boundaries[ -1 ] = times[ -1 ] + 0.01


	norm_factor_a = np.min( values )
	norm_factor_b = np.max( values ) - np.min( values ) 
	normalized_values = ( values - norm_factor_a ) / norm_factor_b
 

	#cdef np.ndarray[ np.float64_t, ndim=1 ] c_input_times = times
	#cdef np.ndarray[ np.float64_t, ndim=1 ] c_input_values = values
	#cdef np.ndarray[ np.float64_t, ndim=1 ] c_boundaries = boundaries
	cdef vector[double] c_input_times = times
	cdef vector[double] c_input_values = normalized_values
	cdef vector[double] c_boundaries = boundaries
	#c_boundaries.push_back( 5 ) 
	#print( c_boundaries )
	#print( c_boundaries.size() )
	cdef int c_init_bounds = init_bounds
	cdef double c_weight_slope = weight_slope
	cdef double c_weight_offset = weight_offset
	cdef double c_weight_tau = weight_tau
	cdef double c_weight_lambda = weight_lambda
	cdef double c_delta_slope = delta_slope
	cdef double c_delta_offset = delta_offset
	cdef double c_delta_tau = delta_tau
	cdef double c_delta_boundary = delta_boundary
	cdef double c_mean_slope = mean_slope
	#cdef double c_mean_offset
	cdef double c_mean_tau = mean_tau
	cdef int c_max_iterations = max_iterations
	cdef int c_max_cost_evaluations = max_cost_evaluations
	cdef double c_rho_end = rho_end
	cdef bool c_use_early_stopping = use_early_stopping
	cdef double c_epsilon = epsilon
	cdef int c_patience = patience
	cdef FitData fit_results
	fit_results = estimate_targets(
						#&c_input_times[0],
						#&c_input_values[0],
						#&c_boundaries[0],
						c_input_times,
						c_input_values,
						c_boundaries,
						#c_init_bounds,
						c_weight_slope,
						c_weight_offset,
						c_weight_tau,
						c_weight_lambda,
						c_delta_slope,
						c_delta_offset,
						c_delta_tau,
						c_delta_boundary,
						c_mean_slope,
						#c_mean_offset,
						c_mean_tau,
						c_max_iterations,
						c_max_cost_evaluations,
						c_rho_end,
						c_use_early_stopping,
						c_epsilon,
						c_patience,
					)
	#print( 'res targets: ------------------------')
	#print( fit_results.res_targets )
	out_targets = [
		tg.Target(
			onset_time = fit_results.res_boundaries.at( i ),
			duration = target['duration'],
			slope = target['slope'] * norm_factor_b,
			offset = target['offset'] * norm_factor_b + norm_factor_a, 
			time_constant = target['tau'] / 1000, # normalization because tau is in [ms] in c++ code 
		) for i, target in enumerate( fit_results.res_targets )
	]
	out_targets[ 0 ].onset_state = fit_results.res_onset * norm_factor_b + norm_factor_a
	tgs = tg.Target_Sequence( targets = out_targets, name = 'Joint Optimization' )
	out_trajectory = tgs.get_contour()
	fit_info = dict(
		in_times = np.array( times ),
		in_values = values,
		in_boundaries = boundaries,
		par_init_bounds = init_bounds,
		par_weight_slope = weight_slope,
		par_weight_offset = weight_offset,
		par_weight_tau = weight_tau,
		par_weight_lambda = weight_lambda,
		par_delta_slope = delta_slope,
		par_delta_offset = delta_offset,
		par_delta_tau = delta_tau,
		par_delta_boundary = delta_boundary,
		par_mean_slope = mean_slope,
		par_mean_tau = mean_tau,
		par_max_iterations = max_iterations,
		par_max_cost_evaluations = max_cost_evaluations,
		par_rho_end = rho_end,
		par_use_early_stopping = use_early_stopping,
		par_epsilon = epsilon,
		par_patience = patience,
		out_targets = out_targets,
		out_boundaries = np.array( fit_results.res_boundaries ),
		out_trajectory = out_trajectory,
		#out_trajectory = np.array( [ [ sample.time, sample.value ] for sample in fit_results.res_trajectory ] ),
		out_ftmp = np.array( fit_results.res_ftmp ),
		out_fmin = fit_results.res_fmin,
		out_rmse = get_rmse( times, values, tgs ),
		out_corr = get_correlation( times, values, tgs ),
		out_time = fit_results.res_time,
	)
	return Fit_Result( **fit_info )
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def fit_sequentially( times, 
	                  values, 
	                  boundaries = [], 
	                  init_bounds: int = 9,
	                  delta_boundary: int = 0, 
	                  window_length: int = 3, 
	                  hop_length: int = 1, 
	                  n_passes: int = 1,
	                  underflow_threshold: float = None,
	                  show_plot = True,
	                  **kwargs 
	                  ):
	if boundaries == [] and init_bounds != 0:
		step =  ( times[ -1 ] - times[ 0 ] ) / ( init_bounds - 1 )
		boundaries = np.array( [ times[ 0 ] + x * step for x in range( 0 , init_bounds ) ] )
	if window_length >= len( boundaries ):
		log.warning( 'window_length must be smaller than number of boundaries for sequential fit! Proceed with global fit now.' )
		return fit(
			times = times,
			values = values,
			boundaries = boundaries,
			delta_boundary = delta_boundary,
			**kwargs,
		)
	if delta_boundary == 0 and n_passes != 1:
		log.warning( 'Passed argument n_passes = {} is deactivated because boundaries are not optimized (delta_boundary = 0).'.format( n_passes ) )
		n_passes = 1
	if delta_boundary > 0 and n_passes == 1:
		log.warning( 'Passed argument n_passes = 1 does not work with argument delta_boundary > 0). Setting n_passes = 2 now.' )
		n_passes = 2
	#n_windows = int( ( ( len( boundaries ) - 1 - window_length ) / hop_length ) + 1 )
	# windows = [ Fit_Window( times, values, boundaries, queue_position = x, 
	# 	window_length= window_length, hop_length=hop_length, **kwargs ) for x in range( 0, n_windows ) ]
	# seq_fit = Sequential_Fit( windows, times, values )
	# if show_plot:
	# 	seq_fit.plot()

	windows_list = []
	for index in range( 0, n_passes ):
		if index == n_passes - 1:
			delta_boundary = 0
		n_windows = int( ( ( len( boundaries ) - 1 - window_length ) / hop_length ) + 1 )
		print( 'Iteration nr: {}'.format( index ) )
		print( len(boundaries) )
		print( 'delta bounds: {}'.format(delta_boundary) )

		windows = [
			Fit_Window(
				times = times,
				values = values,
				boundaries = boundaries,
				delta_boundary = delta_boundary,
				queue_position = x, 
				window_length= window_length,
				hop_length=hop_length,
				**kwargs
			) for x in range( 0, n_windows )
		]
		seq_fit = Sequential_Fit( windows, times, values )
		if show_plot:
			seq_fit.plot()
		#seq_fit = Sequential_Fit( windows, times, values )
		if ( delta_boundary > 0 ) and ( index < n_passes - 1 ):
			boundaries = get_optimized_boundaries( windows, window_length, hop_length, underflow_threshold )
			print( 'New boundaries: {}'.format( boundaries ) )
		windows_list.append( windows )

	out_boundaries = boundaries
	#target_list = []
	# for window in windows:
	# 	if window.queue_position == 0:
	# 		target_list.extend( window.fit_result.out_targets[ : 1 + hop_length ] )
	# 	elif window.queue_position < n_windows - 1:
	# 		target_list.append( window.fit_result.out_targets[ hop_length ] )
	# 	elif window.queue_position == n_windows - 1:
	# 		target_list.extend( window.fit_result.out_targets[ 1 : ] )

	out_targets = get_optimized_targets( windows = windows, hop_length = hop_length )
	tgs = tg.Target_Sequence( targets = out_targets, name = 'Sequential fit' )
	#tgs.plot()
	out_targets[ 0 ].onset_state = windows[0].fit_result.out_targets[ 0 ].onset_state
	out_trajectory = tgs.get_contour()
	fit_info = dict(
		in_times = np.array( times ),
		in_values = values,
		in_boundaries = boundaries,
		par_init_bounds = init_bounds,
		par_weight_slope = windows[0].fit_result.par_weight_slope,
		par_weight_offset = windows[0].fit_result.par_weight_offset,
		par_weight_tau = windows[0].fit_result.par_weight_tau,
		par_weight_lambda = windows[0].fit_result.par_weight_lambda,
		par_delta_slope = windows[0].fit_result.par_delta_slope,
		par_delta_offset = windows[0].fit_result.par_delta_offset,
		par_delta_tau = windows[0].fit_result.par_delta_tau,
		par_delta_boundary = windows[0].fit_result.par_delta_boundary,
		par_mean_slope = windows[0].fit_result.par_mean_slope,
		par_mean_tau = windows[0].fit_result.par_mean_tau,
		par_max_iterations = windows[0].fit_result.par_max_iterations,
		par_max_cost_evaluations = windows[0].fit_result.par_max_cost_evaluations,
		par_rho_end = windows[0].fit_result.par_rho_end,
		par_use_early_stopping = windows[0].fit_result.par_use_early_stopping,
		par_epsilon = windows[0].fit_result.par_epsilon,
		par_patience = windows[0].fit_result.par_patience,
		out_targets = out_targets,
		out_boundaries = np.array( out_boundaries ),
		out_trajectory = out_trajectory,
		out_ftmp = np.array( [ window.fit_result.out_ftmp for window in windows ] ),
		out_fmin = np.mean( [ window.fit_result.out_fmin for window in windows ] ),
		out_rmse = get_rmse( times, values, tgs ),
		out_corr = get_correlation( times, values, tgs ),
		out_time = np.sum( [ window.fit_result.out_time for windows in windows_list for window in windows ] ),
	)
	return Fit_Result( **fit_info )
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def get_optimized_boundaries( windows, window_length, hop_length, underflow_threshold = 0.010 ):
	boundary_list = []
	n_targets = windows[-1].target_positions[-1]# + 1
	print( n_targets )
	target_index = [ x for x in range(0, window_length) ]
	for index in range( 0, hop_length +1 ):
		boundary_list.append( windows[0].fit_result.out_targets[ target_index[ index % len( target_index ) ] ].onset_time )
	for index in range( hop_length + 1, n_targets - 1 ):
		tmp_targets = []
		for window in windows:
			if ( index in window.target_positions ) and ( index != window.target_positions[0] ):
				tmp_targets.append( window.fit_result.out_targets[ target_index[ index % len( target_index ) ] ] )
		onset_time_mean = np.mean( [ x.onset_time for x in tmp_targets ] )
		boundary_list.append( onset_time_mean )
	#for index in range( n_targets - 1, n_targets ):
	boundary_list.append( windows[-1].fit_result.out_targets[ -1 ].onset_time )
	boundary_list.append( windows[-1].fit_result.out_targets[ -1 ].offset_time )
	boundary_list = sorted( boundary_list )
	if underflow_threshold != None:
		boundaries = []
		for index, x in enumerate( boundary_list ):
			if index == 0 or ( x - boundaries[ -1 ] ) > underflow_threshold:
				boundaries.append(x)
		boundary_list = boundaries
	return boundary_list
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def get_optimized_targets( windows, hop_length ):
	n_windows = len( windows )
	target_list = []
	for window in windows:
		if window.queue_position == 0:
			target_list.extend( window.fit_result.out_targets[ : 1 + hop_length ] )
		elif window.queue_position < n_windows - 1:
			target_list.append( window.fit_result.out_targets[ hop_length ] )
		elif window.queue_position == n_windows - 1:
			target_list.extend( window.fit_result.out_targets[ 1 : ] )
	return target_list
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def get_correlation( times, values, target_sequence ):
	fitted_values = target_sequence.get_contour( sample_times = times )[ :, 1 ]
	x = values - np.mean( values );
	y = fitted_values - np.mean( fitted_values );
	return ( np.dot(x, y) ) / ( ( np.sqrt( np.sum( x**2 ) ) ) * ( np.sqrt( np.sum( y**2 ) ) ) );
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def get_rmse( times, values, target_sequence ):
	#print( 'len times: {}'.format( len( times ) ) )
	#print( 'len values: {}'.format( len( values ) ) )
	fitted_values = target_sequence.get_contour( sample_times = times )[ :, 1 ]
	#print( 'len fitted values: {}'.format( len( fitted_values ) ) )

	#cnt = target_sequence.get_contour( sample_times = times )
	#for bound in target_sequence.get_boundaries():
	#	plt.axvline( bound )
	# cnt2 = target_sequence.get_contour()
	# target_sequence.plot( plot_contour = False, show = False )
	# plt.plot( cnt2[ :, 0 ], cnt2[ :, 1 ] )
	#plt.scatter( times, values )
	#plt.scatter( cnt[ :, 0 ], cnt[ :, 1 ] )
	#plt.show()
	return np.sqrt( np.mean( ( values - fitted_values )**2 ) )
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	#for window in windows:
	#	for target in window.fit_result.out_targets:
	#		print( 'tg')
	#		print( 'onset: {}, offset: {}, slope: {}, offset: {}, tau: {}'.format( target.onset_time, target.offset_time, target.slope, target.offset, target.tau ) )




	'''

	return




	partial_signals = []
	partial_boundaries = []
	i = 0
	while i + sequence_length < len( boundaries ):
		partial_signal = []
		#for sample in time_signal:
		for time, value in zip( times, values ):
			if time >= boundaries[ i ] and time <= boundaries[ i + sequence_length ]:
				partial_signal.append( [ time, value ] )
		partial_signals.append( partial_signal )
		partial_boundaries.append( [ boundaries[ x ] for x in range( i, i + sequence_length +1 ) ] )
		i += 1
	partial_fits = []
	for signal, boundaries_par in zip( partial_signals, partial_boundaries ):
		signal = np.array( signal )
		partial_fits.append( fit( signal[:,0], signal[:,1], boundaries_par, **kwargs ) )


	tmp_avg = []
	average_factors = [1,2,3,3,3,3,2,1]

	for index, partial_fit in enumerate( partial_fits ):
		print( '-----------fit:-----')
		tgs= []
		#for tgt in range( 0, len(boundaries)-sequence_length ):
		for _ in range( 0, index ):
			tgs.append( tg.Target( 0, 0, 0, 0, 0 ) )
		for target in partial_fit.out_targets:
			print( 'onset: {}, offset: {}, slope: {}, offset: {}, tau: {}'.format( 
				target.onset_time, target.offset_time, target.slope, target.offset, target.tau ) )
			tgs.append( target )
		for _ in range( index + sequence_length, len(boundaries) ):
			tgs.append( tg.Target( 0, 0, 0, 0, 0 ) )
		tmp_avg.append( tgs )
	

	avg_targets = []

	for index in range( 0, len(boundaries)-1 ):
		if index < len(boundaries)-2: 
			avg_target = tg.Target( np.sum( [ x[ index ].onset_time for x in tmp_avg ] ) / average_factors[ index ],
									np.sum( [ x[ index + 1 ].onset_time for x in tmp_avg ] ) / average_factors[ index+1 ],
									np.sum( [ x[ index ].slope for x in tmp_avg ] ) / average_factors[ index ],
									np.sum( [ x[ index ].offset for x in tmp_avg] ) / average_factors[ index ],
									np.sum( [ x[ index ].tau for x in tmp_avg] ) / average_factors[ index ], 
								)
		else:
			avg_target = tg.Target( np.sum( [ x[ index-1 ].offset_time for x in tmp_avg ] ) / average_factors[ index-1 ],
									np.sum( [ x[ index ].offset_time for x in tmp_avg ] ) / average_factors[ index ],
									np.sum( [ x[ index ].slope for x in tmp_avg ] ) / average_factors[ index ],
									np.sum( [ x[ index ].offset for x in tmp_avg] ) / average_factors[ index ],
									np.sum( [ x[ index ].tau for x in tmp_avg] ) / average_factors[ index ], 
								)
		avg_targets.append( avg_target )



	##average_targets = []
	##average_factors = [1,2,3,3,3,3,2,1]
	#avg_targets = []
	#for index in range(0, len( boundaries )-1 ):
	#	tmp_avg = []
	#	for j, partial_fit in enumerate( partial_fits ):
	#		for target in partial_fit.out_targets:
	#			if np.abs( target.onset_time - boundaries[ index ] ) < 0.0001:
	#				print( 'boundary {}: {}'.format( index, boundaries[index] ) )
	#				print( 'fir nr: ', j)
	#				tmp_avg.append( target )
	#	#avg_target = tg.Target( boundaries[ index ], boundaries[ index + 1 ],
	#	#						np.mean( [ x.slope for x in tmp_avg ] ),
	#	#						np.mean( [ x.offset for x in tmp_avg] ),
	#	#						np.mean( [ x.tau for x in tmp_avg] ), 
	#	#					)
	#	#x = tmp_avg[ int( len(tmp_avg) / 2 ) ]
	#	#avg_target = tg.Target( boundaries[ index ], boundaries[ index + 1 ],
	#	#						x.slope,
	#	#						x.offset,
	#	#						x.tau, 
	#	#					)
	#	avg_target = tg.Target( np.mean( [ x.onset_time for x in tmp_avg ] ),
	#							np.mean( [ x.offset_time for x in tmp_avg ] ),
	#							np.mean( [ x.slope for x in tmp_avg ] ),
	#							np.mean( [ x.offset for x in tmp_avg] ),
	#							np.mean( [ x.tau for x in tmp_avg] ), 
	#						)
	#	avg_targets.append( avg_target )

	'''
	'''
	avg_targets = target_sequence
	print( avg_targets )
	partial_fits = [ window.fit_result for window in windows ]

	print( avg_targets )
	for target in avg_targets:
		print( 'onset: {}, offset: {}, slope: {}, offset: {}, tau: {}'.format( target.onset_time, target.offset_time, target.slope, target.offset, target.tau ) )
	tam = tg.Target_Approximation_Model()
	cnt = tam.response( avg_targets,  onset_state = values[0] )


	#for partial_fit in partial_fits:
		#plt.plot( partial_fit.out_trajectory[:,0], partial_fit.out_trajectory[:,1] )
	plt.plot( cnt[:,0], cnt[:,1] )
	for tgt in avg_targets:
		plt.axvline( tgt.onset_time )
	plt.axvline( avg_targets[-1].offset_time )
	plt.scatter( times, values, c = 'black' )
	plt.plot( partial_fits[0].out_trajectory[:,0], partial_fits[0].out_trajectory[:,1] )
	plt.plot( partial_fits[-1].out_trajectory[:,0], partial_fits[-1].out_trajectory[:,1] )
	plt.show()
	'''
	#return
#---------------------------------------------------------------------------------------------------------------------------------------------------#
#def get_partial_fits()
#####################################################################################################################################################


__version__ = '0.4'

#---------------------------------------------------------------------------------------------------------------------------------------------------#
#####################################################################################################################################################
#
#	Import user functions:
#
#####################################################################################################################################################
#---------------------------------------------------------------------------------------------------------------------------------------------------#
from VocalTractLab.VocalTractLabApi import automatic_calculation_of_TRX_and_TRY
from VocalTractLab.VocalTractLabApi import get_version
from VocalTractLab.VocalTractLabApi import get_constants
from VocalTractLab.VocalTractLabApi import get_param_info
from VocalTractLab.VocalTractLabApi import get_shape
from VocalTractLab.VocalTractLabApi import get_shapes
from VocalTractLab.VocalTractLabApi import load_speaker_file
from VocalTractLab.VocalTractLabApi import get_gestural_score_audio_duration
from VocalTractLab.VocalTractLabApi import gestural_score_to_audio
from VocalTractLab.VocalTractLabApi import gestural_score_to_tract_sequence
from VocalTractLab.VocalTractLabApi import segment_sequence_to_gestural_score
from VocalTractLab.VocalTractLabApi import tract_sequence_to_audio
from VocalTractLab.VocalTractLabApi import motor_sequence_to_spectrogram
from VocalTractLab.VocalTractLabApi import motor_sequence_to_melspectrogram
from VocalTractLab.VocalTractLabApi import tract_sequence_to_limited_tract_sequence
from VocalTractLab.VocalTractLabApi import tract_sequence_to_svg
from VocalTractLab.VocalTractLabApi import tract_sequence_to_transfer_functions
from VocalTractLab.VocalTractLabApi import tract_sequence_to_tube_states
from VocalTractLab.audio_tools import calculate_spectral_distances
from VocalTractLab.text_to_speech import text_to_speech
from VocalTractLab.text_to_speech import tts
from VocalTractLab.function_tools import load
from VocalTractLab.function_tools import save
from VocalTractLab.phonemes import *
#---------------------------------------------------------------------------------------------------------------------------------------------------#
#####################################################################################################################################################


#---------------------------------------------------------------------------------------------------------------------------------------------------#
#####################################################################################################################################################
#
#	Import classes
#
#####################################################################################################################################################
#---------------------------------------------------------------------------------------------------------------------------------------------------#
from VocalTractLab import audio_tools
from VocalTractLab.frequency_domain import Transfer_Function
from VocalTractLab import function_tools
from VocalTractLab import g2p
from VocalTractLab import plotting_tools
from VocalTractLab import mc_generator
from VocalTractLab.segment_sequence import Segment_Sequence
from VocalTractLab.targets import Target
from VocalTractLab.targets import Target_Sequence
from VocalTractLab.targets import Target_Score
from VocalTractLab.targets import Synchronous_Target_Score
from VocalTractLab.targets import Sub_Glottal_Motor_Score
from VocalTractLab.targets import Supra_Glottal_Motor_Score
from VocalTractLab.targets import Motor_Score
from VocalTractLab import target_estimation
from VocalTractLab.tract_sequence import Sub_Glottal_Sequence
from VocalTractLab.tract_sequence import Supra_Glottal_Sequence
from VocalTractLab.tract_sequence import Motor_Sequence
from VocalTractLab.tube_states import Tube_State
#---------------------------------------------------------------------------------------------------------------------------------------------------#
#####################################################################################################################################################



#####################################################################################################################################################
#---------------------------------------------------------------------------------------------------------------------------------------------------#
#	- This file is a part of the VocalTractLab Python module PyVTL, see https://github.com/paul-krug/VocalTractLab
#---------------------------------------------------------------------------------------------------------------------------------------------------#
#
#	- Copyright (C) 2021, Paul Konstantin Krug, Dresden, Germany
#	- https://github.com/paul-krug/VocalTractLab
#	- Author: Paul Konstantin Krug, TU Dresden
#
#	- License info:
#
#		This program is free software: you can redistribute it and/or modify
#		it under the terms of the GNU General Public License as published by
#		the Free Software Foundation, either version 3 of the License, or
#		(at your option) any later version.
#		
#		This program is distributed in the hope that it will be useful,
#		but WITHOUT ANY WARRANTY; without even the implied warranty of
#		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#		GNU General Public License for more details.
#		
#		You should have received a copy of the GNU General Public License
#		along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#---------------------------------------------------------------------------------------------------------------------------------------------------#
#####################################################################################################################################################
#---------------------------------------------------------------------------------------------------------------------------------------------------#

#---------------------------------------------------------------------------------------------------------------------------------------------------#
#####################################################################################################################################################
#---------------------------------------------------------------------------------------------------------------------------------------------------#
# Load essential packages:
import os
import soundfile
import librosa
import numpy as np
import pandas as pd
import itertools
from copy import deepcopy


from scipy.spatial.distance import canberra
from scipy.spatial.distance import cityblock
from scipy.spatial.distance import correlation
from scipy.spatial.distance import euclidean
from scipy.spatial.distance import sqeuclidean

from VocalTractLab.multiprocessing_tools import _run_multiprocessing
from VocalTractLab.function_tools import check_if_input_lists_are_valid
from VocalTractLab.function_tools import make_output_path
from VocalTractLab.function_tools import save
#---------------------------------------------------------------------------------------------------------------------------------------------------#
#####################################################################################################################################################

standard_44kHz_spectrogram_kwargs = dict(
	n_fft = 512, # corresponds to 256/11025 s = 23.2199... ms @ sr = 44100 Hz
	hop_length = 110.25, # corresponds to exactly 2.5 ms @ sr = 44100 Hz
	win_length = 512,
	)

standard_22kHz_spectrogram_kwargs = dict(
	n_fft = 512, # corresponds to 256/11025 s = 23.2199... ms @ sr = 22050 Hz
	hop_length = 55.125, # corresponds to exactly 2.5 ms @ sr = 22050 Hz
	win_length = 512,
	)

standard_16kHz_spectrogram_kwargs = dict(
	n_fft = 256, # corresponds to 2/125 s = 16 ms @ sr = 16000 Hz
	hop_length = 40, # corresponds to exactly 2.5 ms @ sr = 16000 Hz
	win_length = 256,
	)

standard_16kHz_melspectrogram_40_kwargs = dict(
	n_fft = 256, # corresponds to 2/125 s = 16 ms @ sr = 16000 Hz
	hop_length = 40, # corresponds to exactly 2.5 ms @ sr = 16000 Hz
	win_length = 256,
	sr = 16000,
	n_mels = 40,
	fmin = 50,
	fmax = 8000,
	)

standard_16kHz_melspectrogram_80_kwargs = dict(
	n_fft = 256, # corresponds to 2/125 s = 16 ms @ sr = 16000 Hz
	hop_length = 40, # corresponds to exactly 2.5 ms @ sr = 16000 Hz
	win_length = 256,
	sr = 16000,
	n_mels = 80,
	fmin = 50,
	fmax = 8000,
	)

standard_16kHz_mfcc_13_kwargs = dict(
	n_fft = 256, # corresponds to 2/125 s = 16 ms @ sr = 16000 Hz
	hop_length = 40, # corresponds to exactly 2.5 ms @ sr = 16000 Hz
	win_length = 256,
	sr = 16000,
	n_mels = 80,
	fmin = 50,
	fmax = 8000,
	n_mfcc = 13,
	)

standard_16kHz_mfcc_26_kwargs = dict(
	n_fft = 256, # corresponds to 2/125 s = 16 ms @ sr = 16000 Hz
	hop_length = 40, # corresponds to exactly 2.5 ms @ sr = 16000 Hz
	win_length = 256,
	sr = 16000,
	n_mels = 80,
	fmin = 50,
	fmax = 8000,
	n_mfcc = 26,
	)

#---------------------------------------------------------------------------------------------------------------------------------------------------#
def calculate_spectral_distances( reference_list, 
	                              query_list,
	                              #spectral_feature = 'mfcc_13',
	                              distance_metric = sqeuclidean,
								  batch_size = None,
	                              dtw_correction = False,
								  dfw_correction = False,
	                              workers = None,
	                              ):
	reference_list, query_list = check_if_input_lists_are_valid( [ reference_list, query_list ], [ np.ndarray, np.ndarray ] )
	args_list = [ [ [
		reference, query, distance_metric ]
		for query in query_list
		] for reference in reference_list
	]
	spectral_ditstance_list = [ _run_multiprocessing( _calculate_spectral_distance, args, True, workers ) for args in args_list ]
	return spectral_ditstance_list
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def calculate_spectral_features(
	audio_file_path_list,
	spectrogram_file_path_list = None,
	audio_kwargs = dict(
		sr = 16000,
	),
	spectrogram_kwargs = dict(
		n_fft = 256,
		hop_length = 64,
		win_length = 256,
	),
	mel_kwargs = dict(
		sr = 16000,
		n_mels = 40,
		n_fft = 256,
		hop_length = 64,
		win_length = 256,
		fmin = 50,
		fmax = 8000,
	),
	workers = None,
):
	audio_file_path_list, spectrogram_file_path_list = check_if_input_lists_are_valid( 
		[ audio_file_path_list, spectrogram_file_path_list ], [ str, ( str, type(None) ) ] 
	)
	args = [ [ audio_file_path, spectrogram_file_path, audio_kwargs, spectrogram_kwargs, mel_kwargs ]
		for audio_file_path, spectrogram_file_path in itertools.zip_longest( audio_file_path_list, spectrogram_file_path_list ) ]
	_run_multiprocessing( _calculate_spectral_features, args, False, workers )
	return
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def calculate_melspectrograms(
	audio_file_path_list,
	spectrogram_file_path_list = None,
	spectrogram_kwargs = standard_16kHz_spectrogram_kwargs,
	mel_kwargs = standard_16kHz_melspectrogram_80_kwargs,
	workers = None,
):
	audio_file_path_list, spectrogram_file_path_list = check_if_input_lists_are_valid( 
		[ audio_file_path_list, spectrogram_file_path_list ], [ str, ( str, type(None) ) ] 
	)
	args = [ [ audio_file_path, spectrogram_file_path, spectrogram_kwargs, mel_kwargs ]
		for audio_file_path, spectrogram_file_path in itertools.zip_longest( audio_file_path_list, spectrogram_file_path_list ) ]
	_run_multiprocessing( _calculate_melscpectrograms, args, False, workers )
	return
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def get_f0( audio_file_path_list,
	        csv_file_path_list = None,
	        lower_f0_limit = 50,
	        upper_f0_limit = 250,
	        save_file: bool = True,
	        return_data: bool = False,
	        workers = None,
	        ):
	audio_file_path_list, csv_file_path_list = check_if_input_lists_are_valid( [ audio_file_path_list, csv_file_path_list ], 
	                                                                              [ str, ( str, type(None) ) ],
	                                                                            )
	args = [ [audio_file_path, csv_file_path, lower_f0_limit, upper_f0_limit, save_file ]
		for audio_file_path, csv_file_path in itertools.zip_longest( audio_file_path_list, csv_file_path_list ) ]
	f0_data_list = _run_multiprocessing( _get_f0, args, return_data, workers )
	return f0_data_list
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def normalize( audio, normalization = -1 ): #normalisation in dB
	norm_factor = 10**( -1 * normalization * 0.05 ) -1
	norm_max = np.max( np.abs( audio ),axis=0)
	audio /= ( norm_max + ( norm_max * norm_factor ) )
	return audio
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def write( audio, audio_file_path, sr ):
	soundfile.write( audio_file_path, audio, sr )
	return
#---------------------------------------------------------------------------------------------------------------------------------------------------#




#---------------------------------------------------------------------------------------------------------------------------------------------------#
def _calculate_spectral_distance( args ):
	reference, query, distance = args
	tmp = []
	for column in range( 0, query.shape[1] ):
		vec_ref = reference[ :, column ]
		vec_qry = query[ :, column ]
		tmp.append( distance( vec_ref, vec_qry ) )
	return np.mean( tmp )
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def _calculate_spectral_features( args ):
	audio_file_path, spectrogram_file_path, audio_kwargs, spectrogram_kwargs, mel_kwargs  = args
	audio, sr = librosa.load( audio_file_path, sr = audio_kwargs[ 'sr' ] )
	#start_sample = df[ 'frame_start' ] * int( sr / video_options[ 'frame_rate' ] )
	#end_sample = df[ 'frame_end' ] * int( sr / video_options[ 'frame_rate' ] )
	#roi = audio[ start_sample : end_sample ]
	#if len( roi ) > 0:
	#norm_max = np.max( np.abs( roi ) )
	#roi /= ( norm_max + ( norm_max * 0.122018 ) ) # Normalize audio to -1 dB
	spectrogram = np.abs( librosa.stft( y = audio, **spectrogram_kwargs ) )**2
	mel = librosa.feature.melspectrogram( S = spectrogram, **mel_kwargs )
	#if 'mfcc' in features:
	#mfcc = librosa.feature.mfcc( S = librosa.power_to_db( mel ), **mel_settings )
	save( spectrogram, spectrogram_file_path + '.pkl.gzip' )
	save( mel, spectrogram_file_path + '_mel_{}.pkl.gzip'.format( mel_kwargs[ 'n_mels' ] ) )
	return
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def _calculate_melscpectrograms( args ):
	audio_file_path, melspectrogram_file_path, spectrogram_kwargs, melspectrogram_kwargs  = args
	audio, sr = librosa.load( audio_file_path, sr = melspectrogram_kwargs[ 'sr' ] )
	spectrogram = np.abs( librosa.stft( y = audio, **spectrogram_kwargs ) )**2
	melspectrogram = librosa.feature.melspectrogram( S = spectrogram, **melspectrogram_kwargs )
	#if log_scale == True:
	melspectrogram = librosa.power_to_db( melspectrogram )
	if melspectrogram_file_path != None: # TODO: replace with saveFile
		#melspectrogram_file_path = make_output_path( melspectrogram_file_path )#motor_sequence.name.rsplit( '.' )[0] +
		save( melspectrogram, melspectrogram_file_path )
	return
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def _get_f0( args ):
	try:
		import parselmouth
	except Exception:
		raise ModuleNotFoundError( 'Can not load praat-parselmouth module, which is needed for f0 estimation. Is it installed?' )
	audio_file_path, csv_file_path, lower_f0_limit, upper_f0_limit, save_file = args
	if not os.path.exists( audio_file_path ):
		warnings.warn( 'the specified audio file path does not exist: {}. API call will be skipped.'.format( audio_file_path ) )
		return
	pitch = parselmouth.Sound( audio_file_path ).to_pitch()
	pitch_times  = pitch.xs()
	pitch_values = pitch.selected_array[ 'frequency' ]
	pitch_values[ pitch_values == 0 ] = np.nan
	pitch_values[ pitch_values >= upper_f0_limit ] = np.nan
	pitch_values[ pitch_values <= lower_f0_limit ] = np.nan
	#data = np.array( [ pitch_times, pitch_values ] ).T
	df = pd.DataFrame( np.array( [ pitch_times, pitch_values ] ).T, columns = [ 'time', 'f0' ] ).dropna()
	if save_file:
		csv_file_path = make_output_path( csv_file_path, audio_file_path.rsplit( '.' )[0] + '.csv' )
		df.to_csv( csv_file_path )
	return df
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def fade_in_fade_out( data_in, fade_to_sample, fade_from_sample ):
	data_faded_in = fade_in(
		data_in = data_in,
		fade_to_sample = fade_to_sample,
		)
	data_faded_in_out = fade_out(
		data_in= data_faded_in,
		fade_from_sample = fade_from_sample,
		)
	return data_faded_in_out
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def fade_in( data_in, fade_to_sample ):
	data_out = deepcopy( data_in )
	data_out[ : fade_to_sample ] = [
		x * cosine_window(
			x = index,
			a = len( data_in[ : fade_to_sample ] ),
			b = np.pi,
			) 
		for index, x in enumerate(
			data_in[ : fade_to_sample ]
			)
		]
	return data_out
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def fade_out( data_in, fade_from_sample ):
	data_out = deepcopy( data_in )
	data_out[ fade_from_sample : ] = [
		x * cosine_window(
			x = index,
			a = len( data_in[ fade_from_sample : ] )
			)
		for index, x in enumerate(
			data_in[ fade_from_sample : ] )
		]
	return data_out
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def cosine_window( x, a = 1, b = 0 ):
	return  0.5 * np.cos( np.pi * x / a + b ) + 0.5
#---------------------------------------------------------------------------------------------------------------------------------------------------#


# poa = place of articulation (0: Labial, Labiodental; lips, 1: Koronal; tongue front, 2: Dorsal; tongue back, 3: Laryngal; glottis)
# tp: tongue position
# th: tongue height
# lr: lip rounding
import pandas as pd
		
		
phoneme_data = pd.DataFrame( [
	dict(
		phoneme_id = 'v_m_1',
		sampa = [ 'a', 'a:' ],
		phoneme_type = 'vowel',
		phoneme_class = 'monophthong',
		voicing = 1,
		nasality = 0,
		plosivity = 0,
		fricativity = 0,
		laterality = 0,
		place_of_articulation = None,
		tongue_position = 0,
		tongue_height = -3,
		lip_rounding = 0,
	),
	dict(
		phoneme_id = 'v_m_2',
		sampa = [ 'e', 'e:' ],
		phoneme_type = 'vowel',
		phoneme_class = 'monophthong',
		voicing = 1,
		nasality = 0,
		plosivity = 0,
		fricativity = 0,
		laterality = 0,
		place_of_articulation = None,
		tongue_position = -2,
		tongue_height = 1,
		lip_rounding = 0,
	),
	dict(
		phoneme_id = 'v_m_3',
		sampa = [ 'i', 'i:' ],
		phoneme_type = 'vowel',
		phoneme_class = 'monophthong',
		voicing = 1,
		nasality = 0,
		plosivity = 0,
		fricativity = 0,
		laterality = 0,
		place_of_articulation = None,
		tongue_position = -2,
		tongue_height = 3,
		lip_rounding = 0,
	),
	dict(
		phoneme_id = 'v_m_4',
		sampa = [ 'o', 'o:' ],
		phoneme_type = 'vowel',
		phoneme_class = 'monophthong',
		voicing = 1,
		nasality = 0,
		plosivity = 0,
		fricativity = 0,
		laterality = 0,
		place_of_articulation = None,
		tongue_position = 2,
		tongue_height = 1,
		lip_rounding = 1,
	),
	dict(
		phoneme_id = 'v_m_5',
		sampa = [ 'u', 'u:' ],
		phoneme_type = 'vowel',
		phoneme_class = 'monophthong',
		voicing = 1,
		nasality = 0,
		plosivity = 0,
		fricativity = 0,
		laterality = 0,
		place_of_articulation = None,
		tongue_position = 2,
		tongue_height = 3,
		lip_rounding = 1,
	),
	dict(
		phoneme_id = 'v_m_6',
		sampa = [ 'E:' ],
		phoneme_type = 'vowel',
		phoneme_class = 'monophthong',
		voicing = 1,
		nasality = 0,
		plosivity = 0,
		fricativity = 0,
		laterality = 0,
		place_of_articulation = None,
		tongue_position = -2,
		tongue_height = -1,
		lip_rounding = 0,
	),
	dict(
		phoneme_id = 'v_m_7',
		sampa = [ '2', '2:' ],
		phoneme_type = 'vowel',
		phoneme_class = 'monophthong',
		voicing = 1,
		nasality = 0,
		plosivity = 0,
		fricativity = 0,
		laterality = 0,
		place_of_articulation = None,
		tongue_position = -2,
		tongue_height = 1,
		lip_rounding = 1,
	),
	dict(
		phoneme_id = 'v_m_8',
		sampa = [ 'y', 'y:' ],
		phoneme_type = 'vowel',
		phoneme_class = 'monophthong',
		voicing = 1,
		nasality = 0,
		plosivity = 0,
		fricativity = 0,
		laterality = 0,
		place_of_articulation = None,
		tongue_position = -2,
		tongue_height = 3,
		lip_rounding = 1,
	),
	dict(
		phoneme_id = 'v_m_9',
		sampa = [ 'I' ],
		phoneme_type = 'vowel',
		phoneme_class = 'monophthong',
		voicing = 1,
		nasality = 0,
		plosivity = 0,
		fricativity = 0,
		laterality = 0,
		place_of_articulation = None,
		tongue_position = -1,
		tongue_height = 2,
		lip_rounding = 0,
	),
	dict(
		phoneme_id = 'v_m_10',
		sampa = [ 'E' ],
		phoneme_type = 'vowel',
		phoneme_class = 'monophthong',
		voicing = 1,
		nasality = 0,
		plosivity = 0,
		fricativity = 0,
		laterality = 0,
		place_of_articulation = None,
		tongue_position = -2,
		tongue_height = -1,
		lip_rounding = 0,
	),
	dict(
		phoneme_id = 'v_m_11',
		sampa = [ 'O' ],
		phoneme_type = 'vowel',
		phoneme_class = 'monophthong',
		voicing = 1,
		nasality = 0,
		plosivity = 0,
		fricativity = 0,
		laterality = 0,
		place_of_articulation = None,
		tongue_position = 2,
		tongue_height = -1,
		lip_rounding = 1,
	),
	dict(
		phoneme_id = 'v_m_12',
		sampa = [ 'U' ],
		phoneme_type = 'vowel',
		phoneme_class = 'monophthong',
		voicing = 1,
		nasality = 0,
		plosivity = 0,
		fricativity = 0,
		laterality = 0,
		place_of_articulation = None,
		tongue_position = 1,
		tongue_height = 2,
		lip_rounding = 1,
	),
	dict(
		phoneme_id = 'v_m_13',
		sampa = [ '9' ],
		phoneme_type = 'vowel',
		phoneme_class = 'monophthong',
		voicing = 1,
		nasality = 0,
		plosivity = 0,
		fricativity = 0,
		laterality = 0,
		place_of_articulation = None,
		tongue_position = -2,
		tongue_height = -1,
		lip_rounding = 1,
	),
	dict(
		phoneme_id = 'v_m_14',
		sampa = [ 'Y' ],
		phoneme_type = 'vowel',
		phoneme_class = 'monophthong',
		voicing = 1,
		nasality = 0,
		plosivity = 0,
		fricativity = 0,
		laterality = 0,
		place_of_articulation = None,
		tongue_position = -1,
		tongue_height = 2,
		lip_rounding = 1,
	),
	dict(
		phoneme_id = 'v_m_15',
		sampa = [ '@' ],
		phoneme_type = 'vowel',
		phoneme_class = 'monophthong',
		voicing = 1,
		nasality = 0,
		plosivity = 0,
		fricativity = 0,
		laterality = 0,
		place_of_articulation = None,
		tongue_position = 0,
		tongue_height = 0,
		lip_rounding = 0,
	),
	dict(
		phoneme_id = 'v_m_16',
		sampa = [ '6' ],
		phoneme_type = 'vowel',
		phoneme_class = 'monophthong',
		voicing = 1,
		nasality = 0,
		plosivity = 0,
		fricativity = 0,
		laterality = 0,
		place_of_articulation = None,
		tongue_position = 0,
		tongue_height = -2,
		lip_rounding = 0,
	),
	dict(
		phoneme_id = 'v_d_1',
		sampa = [ 'aI' ],
		phoneme_type = 'vowel',
		phoneme_class = 'diphthong',
		voicing = 1,
		nasality = 0,
		plosivity = 0,
		fricativity = 0,
		laterality = 0,
		place_of_articulation = None,
		tongue_position = None,
		tongue_height = None,
		lip_rounding = None,
	),
	dict(
		phoneme_id = 'v_d_2',
		sampa = [ 'aU' ],
		phoneme_type = 'vowel',
		phoneme_class = 'diphthong',
		voicing = 1,
		nasality = 0,
		plosivity = 0,
		fricativity = 0,
		laterality = 0,
		place_of_articulation = None,
		tongue_position = None,
		tongue_height = None,
		lip_rounding = None,
	),
	dict(
		phoneme_id = 'v_d_3',
		sampa = [ 'OY' ],
		phoneme_type = 'vowel',
		phoneme_class = 'diphthong',
		voicing = 1,
		nasality = 0,
		plosivity = 0,
		fricativity = 0,
		laterality = 0,
		place_of_articulation = None,
		tongue_position = None,
		tongue_height = None,
		lip_rounding = None,
	),
	dict(
		phoneme_id = 'v_d_4',
		sampa = [ 'i6', 'i:6' ],
		phoneme_type = 'vowel',
		phoneme_class = 'diphthong',
		voicing = 1,
		nasality = 0,
		plosivity = 0,
		fricativity = 0,
		laterality = 0,
		place_of_articulation = None,
		tongue_position = None,
		tongue_height = None,
		lip_rounding = None,
	),
	dict(
		phoneme_id = 'v_d_5',
		sampa = [ 'I6' ],
		phoneme_type = 'vowel',
		phoneme_class = 'diphthong',
		voicing = 1,
		nasality = 0,
		plosivity = 0,
		fricativity = 0,
		laterality = 0,
		place_of_articulation = None,
		tongue_position = None,
		tongue_height = None,
		lip_rounding = None,
	),
	dict(
		phoneme_id = 'v_d_6',
		sampa = [ 'y6', 'y:6' ],
		phoneme_type = 'vowel',
		phoneme_class = 'diphthong',
		voicing = 1,
		nasality = 0,
		plosivity = 0,
		fricativity = 0,
		laterality = 0,
		place_of_articulation = None,
		tongue_position = None,
		tongue_height = None,
		lip_rounding = None,
	),
	dict(
		phoneme_id = 'v_d_7',
		sampa = [ 'Y6' ],
		phoneme_type = 'vowel',
		phoneme_class = 'diphthong',
		voicing = 1,
		nasality = 0,
		plosivity = 0,
		fricativity = 0,
		laterality = 0,
		place_of_articulation = None,
		tongue_position = None,
		tongue_height = None,
		lip_rounding = None,
	),
	dict(
		phoneme_id = 'v_d_8',
		sampa = [ 'e6', 'e:6' ],
		phoneme_type = 'vowel',
		phoneme_class = 'diphthong',
		voicing = 1,
		nasality = 0,
		plosivity = 0,
		fricativity = 0,
		laterality = 0,
		place_of_articulation = None,
		tongue_position = None,
		tongue_height = None,
		lip_rounding = None,
	),
	dict(
		phoneme_id = 'v_d_9',
		sampa = [ 'E6' ],
		phoneme_type = 'vowel',
		phoneme_class = 'diphthong',
		voicing = 1,
		nasality = 0,
		plosivity = 0,
		fricativity = 0,
		laterality = 0,
		place_of_articulation = None,
		tongue_position = None,
		tongue_height = None,
		lip_rounding = None,
	),
	dict(
		phoneme_id = 'v_d_10',
		sampa = [ 'E:6' ],
		phoneme_type = 'vowel',
		phoneme_class = 'diphthong',
		voicing = 1,
		nasality = 0,
		plosivity = 0,
		fricativity = 0,
		laterality = 0,
		place_of_articulation = None,
		tongue_position = None,
		tongue_height = None,
		lip_rounding = None,
	),
	dict(
		phoneme_id = 'v_d_11',
		sampa = [ '26', '2:6' ],
		phoneme_type = 'vowel',
		phoneme_class = 'diphthong',
		voicing = 1,
		nasality = 0,
		plosivity = 0,
		fricativity = 0,
		laterality = 0,
		place_of_articulation = None,
		tongue_position = None,
		tongue_height = None,
		lip_rounding = None,
	),
	dict(
		phoneme_id = 'v_d_12',
		sampa = [ '96' ],
		phoneme_type = 'vowel',
		phoneme_class = 'diphthong',
		voicing = 1,
		nasality = 0,
		plosivity = 0,
		fricativity = 0,
		laterality = 0,
		place_of_articulation = None,
		tongue_position = None,
		tongue_height = None,
		lip_rounding = None,
	),
	dict(
		phoneme_id = 'v_d_13',
		sampa = [ 'a6', 'a:6' ],
		phoneme_type = 'vowel',
		phoneme_class = 'diphthong',
		voicing = 1,
		nasality = 0,
		plosivity = 0,
		fricativity = 0,
		laterality = 0,
		place_of_articulation = None,
		tongue_position = None,
		tongue_height = None,
		lip_rounding = None,
	),
	dict(
		phoneme_id = 'v_d_14',
		sampa = [ 'u6', 'u:6' ],
		phoneme_type = 'vowel',
		phoneme_class = 'diphthong',
		voicing = 1,
		nasality = 0,
		plosivity = 0,
		fricativity = 0,
		laterality = 0,
		place_of_articulation = None,
		tongue_position = None,
		tongue_height = None,
		lip_rounding = None,
	),
	dict(
		phoneme_id = 'v_d_15',
		sampa = [ 'U6' ],
		phoneme_type = 'vowel',
		phoneme_class = 'diphthong',
		voicing = 1,
		nasality = 0,
		plosivity = 0,
		fricativity = 0,
		laterality = 0,
		place_of_articulation = None,
		tongue_position = None,
		tongue_height = None,
		lip_rounding = None,
	),
	dict(
		phoneme_id = 'v_d_16',
		sampa = [ 'o6', 'o:6' ],
		phoneme_type = 'vowel',
		phoneme_class = 'diphthong',
		voicing = 1,
		nasality = 0,
		plosivity = 0,
		fricativity = 0,
		laterality = 0,
		place_of_articulation = None,
		tongue_position = None,
		tongue_height = None,
		lip_rounding = None,
	),
	dict(
		phoneme_id = 'v_d_17',
		sampa = [ 'O6' ],
		phoneme_type = 'vowel',
		phoneme_class = 'diphthong',
		voicing = 1,
		nasality = 0,
		plosivity = 0,
		fricativity = 0,
		laterality = 0,
		place_of_articulation = None,
		tongue_position = None,
		tongue_height = None,
		lip_rounding = None,
	),
	dict(
		phoneme_id = 'c_a_1',
		sampa = [ 'pf' ],
		phoneme_type = 'consonant',
		phoneme_class = 'affricate',
		voicing = 0,
		nasality = 0,
		plosivity = 1,
		fricativity = 1,
		laterality = 0,
		place_of_articulation = 0,
		tongue_position = None,
		tongue_height = None,
		lip_rounding = None,
	),
	dict(
		phoneme_id = 'c_a_2',
		sampa = [ 'ts' ],
		phoneme_type = 'consonant',
		phoneme_class = 'affricate',
		voicing = 0,
		nasality = 0,
		plosivity = 1,
		fricativity = 1,
		laterality = 0,
		place_of_articulation = 1,
		tongue_position = None,
		tongue_height = None,
		lip_rounding = None,
	),
	dict(
		phoneme_id = 'c_a_3',
		sampa = [ 'tS' ],
		phoneme_type = 'consonant',
		phoneme_class = 'affricate',
		voicing = 0,
		nasality = 0,
		plosivity = 1,
		fricativity = 1,
		laterality = 0,
		place_of_articulation = 1,
		tongue_position = None,
		tongue_height = None,
		lip_rounding = None,
	),
	dict(
		phoneme_id = 'c_a_4',
		sampa = [ 'dZ' ],
		phoneme_type = 'consonant',
		phoneme_class = 'affricate',
		voicing = 1,
		nasality = 0,
		plosivity = 1,
		fricativity = 1,
		laterality = 0,
		place_of_articulation = 1,
		tongue_position = None,
		tongue_height = None,
		lip_rounding = None,
	),
	dict(
		phoneme_id = 'c_p_1',
		sampa = [ '?' ],
		phoneme_type = 'consonant',
		phoneme_class = 'plosive',
		voicing = 1,
		nasality = 0,
		plosivity = 1,
		fricativity = 0,
		laterality = 0,
		place_of_articulation = 3,
		tongue_position = None,
		tongue_height = None,
		lip_rounding = None,
	),
	dict(
		phoneme_id = 'c_p_2',
		sampa = [ 'p' ],
		phoneme_type = 'consonant',
		phoneme_class = 'plosive',
		voicing = 0,
		nasality = 0,
		plosivity = 1,
		fricativity = 0,
		laterality = 0,
		place_of_articulation = 0,
		tongue_position = None,
		tongue_height = None,
		lip_rounding = None,
	),
	dict(
		phoneme_id = 'c_p_3',
		sampa = [ 'b' ],
		phoneme_type = 'consonant',
		phoneme_class = 'plosive',
		voicing = 1,
		nasality = 0,
		plosivity = 1,
		fricativity = 0,
		laterality = 0,
		place_of_articulation = 0,
		tongue_position = None,
		tongue_height = None,
		lip_rounding = None,
	),
	dict(
		phoneme_id = 'c_p_4',
		sampa = [ 't' ],
		phoneme_type = 'consonant',
		phoneme_class = 'plosive',
		voicing = 0,
		nasality = 0,
		plosivity = 1,
		fricativity = 0,
		laterality = 0,
		place_of_articulation = 1,
		tongue_position = None,
		tongue_height = None,
		lip_rounding = None,
	),
	dict(
		phoneme_id = 'c_p_5',
		sampa = [ 'd' ],
		phoneme_type = 'consonant',
		phoneme_class = 'plosive',
		voicing = 1,
		nasality = 0,
		plosivity = 1,
		fricativity = 0,
		laterality = 0,
		place_of_articulation = 1,
		tongue_position = None,
		tongue_height = None,
		lip_rounding = None,
	),
	dict(
		phoneme_id = 'c_p_6',
		sampa = [ 'k' ],
		phoneme_type = 'consonant',
		phoneme_class = 'plosive',
		voicing = 0,
		nasality = 0,
		plosivity = 1,
		fricativity = 0,
		laterality = 0,
		place_of_articulation = 2,
		tongue_position = None,
		tongue_height = None,
		lip_rounding = None,
	),
	dict(
		phoneme_id = 'c_p_7',
		sampa = [ 'g' ],
		phoneme_type = 'consonant',
		phoneme_class = 'plosive',
		voicing = 1,
		nasality = 0,
		plosivity = 1,
		fricativity = 0,
		laterality = 0,
		place_of_articulation = 2,
		tongue_position = None,
		tongue_height = None,
		lip_rounding = None,
	),
	dict(
		phoneme_id = 'c_f_1',
		sampa = [ 'f' ],
		phoneme_type = 'consonant',
		phoneme_class = 'fricative',
		voicing = 0,
		nasality = 0,
		plosivity = 0,
		fricativity = 1,
		laterality = 0,
		place_of_articulation = 0,
		tongue_position = None,
		tongue_height = None,
		lip_rounding = None,
	),
	dict(
		phoneme_id = 'c_f_2',
		sampa = [ 'v' ],
		phoneme_type = 'consonant',
		phoneme_class = 'fricative',
		voicing = 1,
		nasality = 0,
		plosivity = 0,
		fricativity = 1,
		laterality = 0,
		place_of_articulation = 0,
		tongue_position = None,
		tongue_height = None,
		lip_rounding = None,
	),
	dict(
		phoneme_id = 'c_f_3',
		sampa = [ 'T' ],
		phoneme_type = 'consonant',
		phoneme_class = 'fricative',
		voicing = 0,
		nasality = 0,
		plosivity = 0,
		fricativity = 1,
		laterality = 0,
		place_of_articulation = 1,
		tongue_position = None,
		tongue_height = None,
		lip_rounding = None,
	),
	dict(
		phoneme_id = 'c_f_4',
		sampa = [ 'D' ],
		phoneme_type = 'consonant',
		phoneme_class = 'fricative',
		voicing = 1,
		nasality = 0,
		plosivity = 0,
		fricativity = 1,
		laterality = 0,
		place_of_articulation = 1,
		tongue_position = None,
		tongue_height = None,
		lip_rounding = None,
	),
	dict(
		phoneme_id = 'c_f_5',
		sampa = [ 's' ],
		phoneme_type = 'consonant',
		phoneme_class = 'fricative',
		voicing = 0,
		nasality = 0,
		plosivity = 0,
		fricativity = 1,
		laterality = 0,
		place_of_articulation = 1,
		tongue_position = None,
		tongue_height = None,
		lip_rounding = None,
	),
	dict(
		phoneme_id = 'c_f_6',
		sampa = [ 'z' ],
		phoneme_type = 'consonant',
		phoneme_class = 'fricative',
		voicing = 1,
		nasality = 0,
		plosivity = 0,
		fricativity = 1,
		laterality = 0,
		place_of_articulation = 1,
		tongue_position = None,
		tongue_height = None,
		lip_rounding = None,
	),
	dict(
		phoneme_id = 'c_f_7',
		sampa = [ 'S' ],
		phoneme_type = 'consonant',
		phoneme_class = 'fricative',
		voicing = 0,
		nasality = 0,
		plosivity = 0,
		fricativity = 1,
		laterality = 0,
		place_of_articulation = 1,
		tongue_position = None,
		tongue_height = None,
		lip_rounding = None,
	),
	dict(
		phoneme_id = 'c_f_8',
		sampa = [ 'Z' ],
		phoneme_type = 'consonant',
		phoneme_class = 'fricative',
		voicing = 1,
		nasality = 0,
		plosivity = 0,
		fricativity = 1,
		laterality = 0,
		place_of_articulation = 1,
		tongue_position = None,
		tongue_height = None,
		lip_rounding = None,
	),
	dict(
		phoneme_id = 'c_f_9',
		sampa = [ 'C' ],
		phoneme_type = 'consonant',
		phoneme_class = 'fricative',
		voicing = 0,
		nasality = 0,
		plosivity = 0,
		fricativity = 1,
		laterality = 0,
		place_of_articulation = 2,
		tongue_position = None,
		tongue_height = None,
		lip_rounding = None,
	),
	dict(
		phoneme_id = 'c_f_10',
		sampa = [ 'j' ],
		phoneme_type = 'consonant',
		phoneme_class = 'fricative',
		voicing = 1,
		nasality = 0,
		plosivity = 0,
		fricativity = 1,
		laterality = 0,
		place_of_articulation = 2,
		tongue_position = None,
		tongue_height = None,
		lip_rounding = None,
	),
	dict(
		phoneme_id = 'c_f_11',
		sampa = [ 'x' ],
		phoneme_type = 'consonant',
		phoneme_class = 'fricative',
		voicing = 0,
		nasality = 0,
		plosivity = 0,
		fricativity = 1,
		laterality = 0,
		place_of_articulation = 2,
		tongue_position = None,
		tongue_height = None,
		lip_rounding = None,
	),
	dict(
		phoneme_id = 'c_f_12',
		sampa = [ 'r' ], # currently r and R produce the same acoustic in VTL
		phoneme_type = 'consonant',
		phoneme_class = 'fricative',
		voicing = 1,
		nasality = 0,
		plosivity = 0,
		fricativity = 1,
		laterality = 0,
		place_of_articulation = 1,
		tongue_position = None,
		tongue_height = None,
		lip_rounding = None,
	),
	dict(
		phoneme_id = 'c_f_13',
		sampa = [ 'R' ],
		phoneme_type = 'consonant',
		phoneme_class = 'fricative',
		voicing = 1,
		nasality = 0,
		plosivity = 0,
		fricativity = 1,
		laterality = 0,
		place_of_articulation = 2,
		tongue_position = None,
		tongue_height = None,
		lip_rounding = None,
	),
	dict(
		phoneme_id = 'c_f_14',
		sampa = [ 'h' ],
		phoneme_type = 'consonant',
		phoneme_class = 'fricative',
		voicing = 0,
		nasality = 0,
		plosivity = 0,
		fricativity = 1,
		laterality = 0,
		place_of_articulation = 3,
		tongue_position = None,
		tongue_height = None,
		lip_rounding = None,
	),
	dict(
		phoneme_id = 'c_n_1',
		sampa = [ 'm' ],
		phoneme_type = 'consonant',
		phoneme_class = 'nasal',
		voicing = 1,
		nasality = 1,
		plosivity = 0,
		fricativity = 0,
		laterality = 0,
		place_of_articulation = 0,
		tongue_position = None,
		tongue_height = None,
		lip_rounding = None,
	),
	dict(
		phoneme_id = 'c_n_2',
		sampa = [ 'n' ],
		phoneme_type = 'consonant',
		phoneme_class = 'nasal',
		voicing = 1,
		nasality = 1,
		plosivity = 0,
		fricativity = 0,
		laterality = 0,
		place_of_articulation = 1,
		tongue_position = None,
		tongue_height = None,
		lip_rounding = None,
	),
	dict(
		phoneme_id = 'c_n_3',
		sampa = [ 'N' ],
		phoneme_type = 'consonant',
		phoneme_class = 'nasal',
		voicing = 1,
		nasality = 1,
		plosivity = 0,
		fricativity = 0,
		laterality = 0,
		place_of_articulation = 2,
		tongue_position = None,
		tongue_height = None,
		lip_rounding = None,
	),
	dict(
		phoneme_id = 'c_l_1',
		sampa = [ 'l' ],
		phoneme_type = 'consonant',
		phoneme_class = 'lateral',
		voicing = 1,
		nasality = 0,
		plosivity = 0,
		fricativity = 0,
		laterality = 1,
		place_of_articulation = 1,
		tongue_position = None,
		tongue_height = None,
		lip_rounding = None,
		),
] )


#####################################################################################################################################################
#---------------------------------------------------------------------------------------------------------------------------------------------------#
#	- This file is a part of the VocalTractLab Python module PyVTL, see https://github.com/paul-krug/VocalTractLab
#---------------------------------------------------------------------------------------------------------------------------------------------------#
#
#	- Copyright (C) 2021, Paul Konstantin Krug, Dresden, Germany
#	- https://github.com/paul-krug/VocalTractLab
#	- Author: Paul Konstantin Krug, TU Dresden
#
#	- License info:
#
#		This program is free software: you can redistribute it and/or modify
#		it under the terms of the GNU General Public License as published by
#		the Free Software Foundation, either version 3 of the License, or
#		(at your option) any later version.
#		
#		This program is distributed in the hope that it will be useful,
#		but WITHOUT ANY WARRANTY; without even the implied warranty of
#		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#		GNU General Public License for more details.
#		
#		You should have received a copy of the GNU General Public License
#		along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#---------------------------------------------------------------------------------------------------------------------------------------------------#
#####################################################################################################################################################
#---------------------------------------------------------------------------------------------------------------------------------------------------#

#---------------------------------------------------------------------------------------------------------------------------------------------------#
#####################################################################################################################################################
#---------------------------------------------------------------------------------------------------------------------------------------------------#
# Load essential packages:
import os
import sys
import pandas as pd
from VocalTractLab.function_tools import check_if_list_is_valid
WORKING_PATH = os.getcwd()
sys.path.append( os.path.join( os.path.dirname(__file__), 'data', 'dictionaries', 'phonecodes' ) )
import phonecodes as pc
#---------------------------------------------------------------------------------------------------------------------------------------------------#
#####################################################################################################################################################


vtl_phonemes = pd.read_csv( os.path.join( os.path.dirname(__file__), 'data', 'vtl_phonemes.txt' ), sep = ',', skiprows = 4 )
#print( vtl_phonemes )

def text_to_phonemes( sentences, phoneme_style = 'vtl', language = 'en' ):
	sentences = check_if_list_is_valid( sentences, str )
	available_styles = [ 'ipa', 'arpabet', 'xsampa', 'vtl', 'vtlsampa', 'vtl_sampa', 'disc', 'callhome' ]
	if phoneme_style not in available_styles:
		raise ValueError( 'The specified phoneme style must be one of the following: {}, not {}!'.format( available_styles, phoneme_style ) )
	elif phoneme_style in [ 'vtl', 'vtlsampa', 'vtl_sampa' ]:
		alphabet = 'xsampa'
	else:
		alphabet = phoneme_style
	if language == 'de':
		out = _de_to_phonemes( sentences, alphabet )
	elif language == 'en':
		direct_out = _en_to_phonemes( sentences, alphabet )
		out = []
		out_phones = []
		tmp = []
		for sentence in direct_out:
			out_phones = []
			tmp = []
			for x in sentence:
				if x not in [ ' ', '.' ]:
					tmp.append( x )
				else:
					vtl_tmp = get_vtl_phonemes( tmp )
					if vtl_tmp != []:
						out_phones.append( vtl_tmp )
					tmp = []
			out.append( out_phones )
		#print( out )
	#if phoneme_style in [ 'vtl', 'vtlsampa', 'vtl_sampa' ]:
	#	out = [ get_vtl_phonemes( x ) for sentence in out for x in sentence ]
	return out

def _de_to_phonemes( sentences, phoneme_style ):
	phonemes_list = [ dictionary_lookup( sentence, language = 'de' ) for sentence in sentences ]
	if phoneme_style == 'arpabet':
		return phonemes_list
	else:
		phonemes_list = [ [ pc.convert( x, 'arpabet', 'ipa', language='en' ) for x in phonemes ] for phonemes in phonemes_list ]
	if phonemes_list != 'ipa':
		phonemes_list = [ [ pc.convert( x, 'ipa', phoneme_style, language='en' ) for x in phonemes ] for phonemes in phonemes_list ]
	return phonemes_list

def _en_to_phonemes( sentences, phoneme_style ):
	from g2p_en import G2p
	g2p = G2p()
	phonemes_list = [ g2p( sentence ) for sentence in sentences ]
	if phoneme_style == 'arpabet':
		return phonemes_list
	else:
		phonemes_list = [ [ pc.convert( x, 'arpabet', 'ipa', language='en' ) for x in phonemes ] for phonemes in phonemes_list ]
	if phonemes_list != 'ipa':
		phonemes_list = [ [ pc.convert( x, 'ipa', phoneme_style, language='en' ) for x in phonemes ] for phonemes in phonemes_list ]
	return phonemes_list



def get_vtl_phonemes( sampa, drop = [ ',','^','_','.','/','\\','"','%','[',']','=',"'",'!' ] ):
	phones = []
	phoneme_list=[]
	#print( sampa)
	for entry in sampa:
		for char in drop:
			entry = entry.replace(char, '')
		phones.append( entry )
	#print('phones: {}'.format( phones ) )
	return phones
	#phonemes = ''.join( phones )
	found= False
	for _phonemes in phones:
		tmp_list=[]
		while len(_phonemes) > 0:
			found= False
			for entry in phonemes.vtl_sampa:
				if found == False and _phonemes.startswith( entry ):
					tmp_list.append( entry )
					_phonemes = _phonemes[ len( entry ): ]
					found = True
					#print(phonemes)
		phoneme_list.append( tmp_list )
	#print(phoneme_list)

	index = 0
	while index < len( phoneme_list )-1:
		if ( phoneme_list[ index ][ -1 ] in phonemes.vtl_sampa_dict[ 'vowels' ] ) and ( phoneme_list[ index + 1 ][ 0 ] in phonemes.vtl_sampa_dict[ 'vowels' ] ):
			phoneme_list.insert( index + 1, '?' )
		index += 1
	phoneme_list_flat = list( itertools.chain( *phoneme_list ) )

	if ( phoneme_list_flat[0] in phonemes.vtl_sampa_dict['vowels'] ) or ( phoneme_list_flat[0] in phonemes.vtl_sampa_dict['diphthongs'] ):
		phoneme_list_flat.insert(0, '?')
	#print( 'List of phonemes: {}'.format( phoneme_list_flat ) )
	return phoneme_list_flat


def lookup( text, remove_punctuation = True, language = 'de' ):
	if remove_punctuation:
		for char in [ ',','^','_','.','/','\\','"','%','[',']','=',"'",'!','?' ]:
			text = text.replace(char, '')
	dictionary = pd.read_csv( 'Dictionaries/de-ipa.txt', sep='\t' )
	dictionary['de'] = dictionary['de'].apply( lambda x: x.lower() )
	#print( dictionary)
	input_words = text.split(' ')
	#input_words = [ x.lower() for x in input_words ]
	input_words = [ x for x in input_words if x not in [ None, '', ' ' ] ]
	input_words = [ x.lower() for x in input_words ]
	print( input_words )
	ipa_words = []
	#ipa_sub_words = []
	#valid_sub_words = []
	for word in input_words:
		occurences = dictionary.loc[ ( dictionary['de'] == word ) ]
		if len( occurences ) > 0:
			ipa_words.append( occurences[0] )
		else:
			ipa_words.append( None )
	return ipa_words